import numpy

from .._common import jitted


@jitted
def normc(ee, nmat):
    """Normalize Haskell or Dunkin vectors."""
    t1 = 0.0
    for i in range(nmat):
        t1 = max(t1, numpy.abs(ee[i]))

    if t1 < 1.0e-40:
        t1 = 1.0

    for i in range(nmat):
        ee[i] /= t1

    ex = numpy.log(t1)
    return ee, ex



import helpers
import numpy
import pytest

import disba


@pytest.mark.parametrize(
    "wave, water_layer, eref",
    [
        ("rayleigh", False, {"ur": 1.352, "uz": 4.702, "tz": -2.372, "tr": 2.038}),
        ("rayleigh", True, {"ur": 1.155, "uz": 4.965, "tz": -2.260, "tr": 1.638}),
        ("love", False, {"uu": 2.421, "tt": -1.857}),
        ("love", True, {"uu": 1.938, "tt": -2.507}),
    ],
)
def test_eigen(wave, water_layer, eref):
    velocity_model = helpers.velocity_model(5, water_layer=water_layer)

    eigf = disba.EigenFunction(*velocity_model, algorithm="dunkin")
    eig = eigf(10.0, 0, wave)

    for k, v in eref.items():
        assert numpy.allclose(v, getattr(eig, k).sum(), atol=1.0e-3)



from .__about__ import __version__
from ._cps import srfker96, surf96, swegn96
from ._dispersion import DispersionCurve, GroupDispersion, PhaseDispersion
from ._eigen import EigenFunction, LoveEigen, RayleighEigen
from ._ellipticity import Ellipticity, RayleighEllipticity
from ._exception import DispersionError
from ._helpers import depthplot
from ._sensitivity import (
    EllipticitySensitivity,
    GroupSensitivity,
    PhaseSensitivity,
    SensitivityKernel,
)

__all__ = [
    "srfker96",
    "surf96",
    "swegn96",
    "DispersionError",
    "DispersionCurve",
    "PhaseDispersion",
    "GroupDispersion",
    "SensitivityKernel",
    "PhaseSensitivity",
    "GroupSensitivity",
    "EllipticitySensitivity",
    "LoveEigen",
    "RayleighEigen",
    "EigenFunction",
    "RayleighEllipticity",
    "Ellipticity",
    "depthplot",
    "__version__",
]



from collections import namedtuple

import numpy

from ._base import Base
from ._common import ifunc
from ._cps import swegn96
from ._exception import DispersionError

__all__ = [
    "RayleighEllipticity",
    "Ellipticity",
]


RayleighEllipticity = namedtuple(
    "RayleighEllipticity", ("period", "ellipticity", "mode")
)


class Ellipticity(Base):
    def __init__(
        self, thickness, velocity_p, velocity_s, density, algorithm="dunkin", dc=0.005,
    ):
        """
        Ellipticity class (only Rayleigh-wave).

        Parameters
        ----------
        thickness : array_like
            Layer thickness (in km).
        velocity_p : array_like
            Layer P-wave velocity (in km/s).
        velocity_s : array_like
            Layer S-wave velocity (in km/s).
        density : array_like
            Layer density (in g/cm3).
        algorithm : str {'dunkin', 'fast-delta'}, optional, default 'dunkin'
            Algorithm to use for computation of Rayleigh-wave dispersion:
             - 'dunkin': Dunkin's matrix (adapted from surf96),
             - 'fast-delta': fast delta matrix (after Buchen and Ben-Hador, 1996).
        dc : scalar, optional, default 0.005
            Phase velocity increment for root finding.

        """
        super().__init__(thickness, velocity_p, velocity_s, density, algorithm, dc)

    def __call__(self, t, mode=0):
        """
        Compute Rayleigh-wave ellipticity for input period axis and mode.

        Parameters
        ----------
        t : array_like
            Periods (in s).
        mode : int, optional, default 0
            Mode number (0 if fundamental).

        Returns
        -------
        :class:`disba.RayleighEllipticity`
            Rayleigh-wave ellipticity as a namedtuple (period, ellipticity, mode).

        """
        ell = []
        for i, tt in enumerate(t):
            try:
                eig = swegn96(
                    tt,
                    self._thickness,
                    self._velocity_p,
                    self._velocity_s,
                    self._density,
                    mode,
                    ifunc[self._algorithm]["rayleigh"],
                    self._dc,
                )[:, :2]
                ell.append(eig[0, 0] / eig[0, 1])
            except DispersionError:
                i -= 1
                break

        return RayleighEllipticity(t[: i + 1], numpy.array(ell), mode)



import numpy

from ._common import jitted

__all__ = [
    "depthplot",
]


@jitted
def is_sorted(t):
    """Check if array is sorted."""
    for i in range(t.size - 1):
        if t[i + 1] < t[i]:
            return False

    return True


@jitted
def resample(thickness, velocity_p, velocity_s, density, dz):
    """Resample velocity model."""
    mmax = len(thickness)

    sizes = numpy.empty(mmax, dtype=numpy.int32)
    for i in range(mmax):
        sizes[i] = numpy.ceil(thickness[i] / dz) if thickness[i] > dz else 1

    size = sizes.sum()
    d = numpy.empty(size, dtype=numpy.float64)
    a = numpy.empty(size, dtype=numpy.float64)
    b = numpy.empty(size, dtype=numpy.float64)
    rho = numpy.empty(size, dtype=numpy.float64)

    j = 0
    for i in range(mmax):
        dzi = thickness[i] / sizes[i]
        for _ in range(sizes[i]):
            d[j] = dzi
            a[j] = velocity_p[i]
            b[j] = velocity_s[i]
            rho[j] = density[i]
            j += 1

    return d, a, b, rho


def depthplot(thickness, parameter, zmax=None, plot_args=None, ax=None):
    """
    Plot parameter against depth.

    Parameters
    ----------
    thickness : array_like
        Layer's thickness.
    parameter : array_like
        Parameter to plot against depth.
    zmax : scalar or None, optional, default None
        Depth of last data point.
    plot_args : dict or None, optional, default None
        Plot arguments passed to :func:`matplotlib.pyplot.plot`.
    ax : matplotlib.pyplot.Axes or None, optional, default None
        Matplotlib axes. If `None`, use current active plot.

    """
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("depthplot requires matplotlib to be installed")

    x = parameter
    z = numpy.cumsum(thickness)
    n = z.size

    if len(parameter) != n:
        raise ValueError()

    # Plot arguments
    plot_args = plot_args if plot_args is not None else {}
    _plot_args = {
        "color": "black",
        "linewidth": 2,
    }
    _plot_args.update(plot_args)

    # Determine zmax
    if zmax is None:
        tmp = numpy.array(thickness)
        tmp[-1] = tmp[:-1].min()
        zmax = tmp.sum()

    # Build layered model
    xin = numpy.empty(2 * n)
    xin[1::2] = x
    xin[2::2] = x[1:]
    xin[0] = xin[1]

    zin = numpy.zeros_like(xin)
    zin[1:-1:2] = z[:-1]
    zin[2::2] = z[:-1]
    zin[-1] = max(z[-1], zmax)

    # Plot
    plot = getattr(plt if ax is None else ax, "plot")
    plot(xin, zin, **_plot_args)

    ax = ax if ax is not None else plt.gca()
    ax.set_ylim(zmax, zin.min())



import numpy

from .._common import jitted
from ._surf96 import surf96

__all__ = [
    "srfker96",
]


@jitted
def srfker96(
    t, d, a, b, rho, mode=0, itype=0, ifunc=2, ipar=2, dc=0.005, dt=0.025, dp=0.025
):
    """
    Get phase or group velocity sensitivity kernel.

    Parameters
    ----------
    t : scalar
        Period (in s).
    d : array_like
        Layer thickness (in km).
    a : array_like
        Layer P-wave velocity (in km/s).
    b : array_like
        Layer S-wave velocity (in km/s).
    rho : array_like
        Layer density (in g/cm3).
    mode : int, optional, default 0
        Mode number (0 if fundamental).
    itype : int, optional, default 0
        Velocity type:
         - 0: phase velocity,
         - 1: group velocity.
    ifunc : int, optional, default 2
        Select wave type and algorithm for period equation:
         - 1: Love-wave (Thomson-Haskell method),
         - 2: Rayleigh-wave (Dunkin's matrix),
         - 3: Rayleigh-wave (fast delta matrix).
    ipar : int, optional, default 2
        Select parameter with respect to which sensitivity kernel is calculated:
         - 0: Layer thickness,
         - 1: P-wave velocity,
         - 2: S-wave velocity,
         - 3: Layer density.
    dc : scalar, optional, default 0.005
        Phase velocity increment for root finding.
    dt : scalar, optional, default 0.025
        Frequency increment (%) for calculating group velocity.
    dp : scalar, optional, default 0.025
        Parameter increment (%) for numerical partial derivatives.

    Returns
    -------
    array_like
        Phase or group velocity sensitivity kernel.

    """
    # Reference velocity
    period = numpy.empty(1, dtype=numpy.float64)
    period[0] = t
    c1 = surf96(period, d, a, b, rho, mode, itype, ifunc, dc, dt)

    # Initialize kernel
    mmax = len(d)
    kernel = numpy.zeros(mmax, dtype=numpy.float64)

    # Love-waves are not sensitive to P-wave
    if not (ipar == 1 and ifunc == 1):
        # Copy parameter array
        par = numpy.empty(mmax, dtype=numpy.float64)
        if ipar == 0:
            for i in range(mmax):
                par[i] = d[i]
        elif ipar == 1:
            for i in range(mmax):
                par[i] = a[i]
        elif ipar == 2:
            for i in range(mmax):
                par[i] = b[i]
        else:
            for i in range(mmax):
                par[i] = rho[i]

        # Ignore top and/or bottom layers depending on inputs
        ibeg = 1 if b[0] < 0.01 and (ipar == 2 or ifunc == 0) else 0
        iend = mmax - 1 if ipar == 0 else mmax

        # Loop over layers
        fac = 1.0 + dp
        for i in range(ibeg, iend):
            tmp = par[i]
            par[i] /= fac

            if ipar == 0:
                c2 = surf96(period, par, a, b, rho, mode, itype, ifunc, dc, dt)
            elif ipar == 1:
                c2 = surf96(period, d, par, b, rho, mode, itype, ifunc, dc, dt)
            elif ipar == 2:
                c2 = surf96(period, d, a, par, rho, mode, itype, ifunc, dc, dt)
            else:
                c2 = surf96(period, d, a, b, par, mode, itype, ifunc, dc, dt)

            kernel[i] = (c2[0] - c1[0]) / (par[i] - tmp)
            par[i] *= fac

    return c1[0], kernel



import helpers
import numpy
import pytest

import disba


@pytest.mark.parametrize("mode, eref", [(0, 14.038), (1, 3.290)])
def test_ellipticity(mode, eref):
    velocity_model = helpers.velocity_model(5)
    t = numpy.logspace(0.0, 1.0, 20)

    ell = disba.Ellipticity(*velocity_model, algorithm="dunkin")
    rel = ell(t, mode=mode)

    assert numpy.allclose(eref, rel.ellipticity.sum(), atol=0.001)



import helpers
import numpy
import pytest

import disba


@pytest.mark.parametrize(
    "mode, wave, algorithm, cref",
    [
        (0, "rayleigh", "dunkin", 11.788),
        (0, "rayleigh", "fast-delta", 11.782),
        (1, "rayleigh", "dunkin", 8.032),
        (1, "rayleigh", "fast-delta", 8.819),
        (0, "love", "dunkin", 12.900),
        (1, "love", "dunkin", 7.295),
    ],
)
def test_phase(mode, wave, algorithm, cref):
    velocity_model = helpers.velocity_model(5)
    t = numpy.logspace(0.0, 1.0, 20)

    pd = disba.PhaseDispersion(*velocity_model, algorithm=algorithm)
    cp = pd(t, mode, wave)

    assert numpy.allclose(cref, cp.velocity.sum(), atol=0.001)


@pytest.mark.parametrize(
    "mode, wave, algorithm, cref",
    [
        (0, "rayleigh", "dunkin", 10.667),
        (0, "rayleigh", "fast-delta", 10.705),
        (1, "rayleigh", "dunkin", 6.931),
        (1, "rayleigh", "fast-delta", 6.914),
        (0, "love", "dunkin", 11.766),
        (1, "love", "dunkin", 6.255),
    ],
)
def test_group(mode, wave, algorithm, cref):
    velocity_model = helpers.velocity_model(5)
    t = numpy.logspace(0.0, 1.0, 20)

    gd = disba.GroupDispersion(*velocity_model, algorithm=algorithm)
    cg = gd(t, mode, wave)

    assert numpy.allclose(cref, cg.velocity.sum(), atol=0.001)


@pytest.mark.parametrize(
    "wave, algorithm, cref",
    [
        ("rayleigh", "dunkin", 11.596),
        ("rayleigh", "fast-delta", 11.581),
        ("love", "dunkin", 14.108),
    ],
)
def test_water_layer(wave, algorithm, cref):
    velocity_model = helpers.velocity_model(5, water_layer=True)
    t = numpy.logspace(0.0, 1.0, 20)

    pd = disba.PhaseDispersion(*velocity_model, algorithm=algorithm)
    cp = pd(t, wave=wave)

    assert numpy.allclose(cref, cp.velocity.sum(), atol=0.001)



"""
Numba implementation of the Fortran program surf96.

This module is not a one-to-one translation from Fortran to Python.
The code has been adapted and optimized for Numba.

..

    COMPUTER PROGRAMS IN SEISMOLOGY
    VOLUME IV

    COPYRIGHT 1986, 1991
    D. R. Russell, R. B. Herrmann
    Department of Earth and Atmospheric Sciences
    Saint Louis University
    221 North Grand Boulevard
    St. Louis, Missouri 63103
    U. S. A.

"""

import numpy

from .._common import jitted
from .._exception import DispersionError
from ._common import normc

__all__ = [
    "surf96",
]


twopi = 2.0 * numpy.pi


@jitted
def dnka(wvno2, gam, gammk, rho, a0, cpcq, cpy, cpz, cqw, cqx, xy, xz, wy, wz, ca):
    """Dunkin's matrix."""
    gamm1 = gam - 1.0
    twgm1 = gam + gamm1
    gmgmk = gam * gammk
    gmgm1 = gam * gamm1
    gm1sq = gamm1 * gamm1

    rho2 = rho * rho
    a0pq = a0 - cpcq
    t = -2.0 * wvno2

    ca[0, 0] = cpcq - 2.0 * gmgm1 * a0pq - gmgmk * xz - wvno2 * gm1sq * wy
    ca[0, 1] = (wvno2 * cpy - cqx) / rho
    ca[0, 2] = -(twgm1 * a0pq + gammk * xz + wvno2 * gamm1 * wy) / rho
    ca[0, 3] = (cpz - wvno2 * cqw) / rho
    ca[0, 4] = -(2.0 * wvno2 * a0pq + xz + wvno2 * wvno2 * wy) / rho2

    ca[1, 0] = (gmgmk * cpz - gm1sq * cqw) * rho
    ca[1, 1] = cpcq
    ca[1, 2] = gammk * cpz - gamm1 * cqw
    ca[1, 3] = -wz
    ca[1, 4] = ca[0, 3]

    ca[3, 0] = (gm1sq * cpy - gmgmk * cqx) * rho
    ca[3, 1] = -xy
    ca[3, 2] = gamm1 * cpy - gammk * cqx
    ca[3, 3] = ca[1, 1]
    ca[3, 4] = ca[0, 1]

    ca[4, 0] = (
        -(2.0 * gmgmk * gm1sq * a0pq + gmgmk * gmgmk * xz + gm1sq * gm1sq * wy) * rho2
    )
    ca[4, 1] = ca[3, 0]
    ca[4, 2] = (
        -(gammk * gamm1 * twgm1 * a0pq + gam * gammk * gammk * xz + gamm1 * gm1sq * wy)
        * rho
    )
    ca[4, 3] = ca[1, 0]
    ca[4, 4] = ca[0, 0]

    ca[2, 0] = t * ca[4, 2]
    ca[2, 1] = t * ca[3, 2]
    ca[2, 2] = a0 + 2.0 * (cpcq - ca[0, 0])
    ca[2, 3] = t * ca[1, 2]
    ca[2, 4] = t * ca[0, 2]

    return ca


@jitted
def var(p, q, ra, rb, wvno, xka, xkb, dpth):
    """Find variables cosP, cosQ, sinP, sinQ..."""
    # Examine P-wave eigenfunctions
    # Checking whether c > vp, c = vp or c < vp
    pex = 0.0
    if wvno < xka:
        sinp = numpy.sin(p)
        w = sinp / ra
        x = -ra * sinp
        cosp = numpy.cos(p)
    elif wvno == xka:
        cosp = 1.0
        w = dpth
        x = 0.0
    elif wvno > xka:
        pex = p
        fac = numpy.exp(-2.0 * p) if p < 16.0 else 0.0
        cosp = (1.0 + fac) * 0.5
        sinp = (1.0 - fac) * 0.5
        w = sinp / ra
        x = ra * sinp

    # Examine S-wave eigenfunctions
    # Checking whether c > vs, c = vs or c < vs
    sex = 0.0
    if wvno < xkb:
        sinq = numpy.sin(q)
        y = sinq / rb
        z = -rb * sinq
        cosq = numpy.cos(q)
    elif wvno == xkb:
        cosq = 1.0
        y = dpth
        z = 0.0
    elif wvno > xkb:
        sex = q
        fac = numpy.exp(-2.0 * q) if q < 16.0 else 0.0
        cosq = (1.0 + fac) * 0.5
        sinq = (1.0 - fac) * 0.5
        y = sinq / rb
        z = rb * sinq

    # Form eigenfunction products for use with compound matrices
    exa = pex + sex
    a0 = numpy.exp(-exa) if exa < 60.0 else 0.0
    cpcq = cosp * cosq
    cpy = cosp * y
    cpz = cosp * z
    cqw = cosq * w
    cqx = cosq * x
    xy = x * y
    xz = x * z
    wy = w * y
    wz = w * z

    qmp = sex - pex
    fac = numpy.exp(qmp) if qmp > -40.0 else 0.0
    cosq *= fac
    y *= fac
    z *= fac

    return w, cosp, a0, cpcq, cpy, cpz, cqw, cqx, xy, xz, wy, wz


@jitted
def dltar1(wvno, omega, d, a, b, rho, llw):
    """Love-wave period equation."""
    beta1 = b[-1]
    rho1 = rho[-1]
    xkb = omega / beta1
    wvnop = wvno + xkb
    wvnom = numpy.abs(wvno - xkb)
    rb = numpy.sqrt(wvnop * wvnom)
    e1 = rho1 * rb
    e2 = 1.0 / (beta1 * beta1)

    for m in range(len(d) - 2, llw, -1):
        beta1 = b[m]
        rho1 = rho[m]
        xmu = rho1 * beta1 * beta1
        xkb = omega / beta1
        wvnop = wvno + xkb
        wvnom = numpy.abs(wvno - xkb)
        rb = numpy.sqrt(wvnop * wvnom)
        q = d[m] * rb

        if wvno < xkb:
            sinq = numpy.sin(q)
            y = sinq / rb
            z = -rb * sinq
            cosq = numpy.cos(q)
        elif wvno == xkb:
            cosq = 1.0
            y = d[m]
            z = 0.0
        else:
            fac = numpy.exp(-2.0 * q) if q < 16.0 else 0.0
            cosq = (1.0 + fac) * 0.5
            sinq = (1.0 - fac) * 0.5
            y = sinq / rb
            z = rb * sinq

        e10 = e1 * cosq + e2 * xmu * z
        e20 = e1 * y / xmu + e2 * cosq
        xnor = numpy.abs(e10)
        ynor = numpy.abs(e20)
        xnor = max(xnor, ynor)
        if xnor < 1.0e-40:
            xnor = 1.0
        e1 = e10 / xnor
        e2 = e20 / xnor

    return e1


@jitted
def dltar4(wvno, omega, d, a, b, rho, llw, ca):
    """Rayleigh-wave period equation."""
    e = numpy.zeros(5, dtype=numpy.float64)
    ee = numpy.zeros(5, dtype=numpy.float64)

    omega = max(omega, 1.0e-4)
    wvno2 = wvno * wvno
    xka = omega / a[-1]
    xkb = omega / b[-1]
    wvnop = wvno + xka
    wvnom = numpy.abs(wvno - xka)
    ra = numpy.sqrt(wvnop * wvnom)
    wvnop = wvno + xkb
    wvnom = numpy.abs(wvno - xkb)
    rb = numpy.sqrt(wvnop * wvnom)
    t = b[-1] / omega

    # E matrix for the bottom half-space
    gammk = 2.0 * t * t
    gam = gammk * wvno2
    gamm1 = gam - 1.0
    rho1 = rho[-1]
    e[0] = rho1 * rho1 * (gamm1 * gamm1 - gam * gammk * ra * rb)
    e[1] = -rho1 * ra
    e[2] = rho1 * (gamm1 - gammk * ra * rb)
    e[3] = rho1 * rb
    e[4] = wvno2 - ra * rb

    # Matrix multiplication from bottom layer upward
    for m in range(len(d) - 2, llw, -1):
        xka = omega / a[m]
        xkb = omega / b[m]
        t = b[m] / omega
        gammk = 2.0 * t * t
        gam = gammk * wvno2
        wvnop = wvno + xka
        wvnom = numpy.abs(wvno - xka)
        ra = numpy.sqrt(wvnop * wvnom)
        wvnop = wvno + xkb
        wvnom = numpy.abs(wvno - xkb)
        rb = numpy.sqrt(wvnop * wvnom)

        dpth = d[m]
        rho1 = rho[m]
        p = ra * dpth
        q = rb * dpth

        # Evaluate cosP, cosQ...
        _, _, a0, cpcq, cpy, cpz, cqw, cqx, xy, xz, wy, wz = var(
            p, q, ra, rb, wvno, xka, xkb, dpth
        )

        # Evaluate Dunkin's matrix
        ca = dnka(
            wvno2, gam, gammk, rho1, a0, cpcq, cpy, cpz, cqw, cqx, xy, xz, wy, wz, ca
        )

        for i in range(5):
            ee[i] = 0.0
            for j in range(5):
                ee[i] += e[j] * ca[j, i]

        ee, _ = normc(ee, 5)
        for i in range(5):
            e[i] = ee[i]

    if llw == 0:
        xka = omega / a[0]
        wvnop = wvno + xka
        wvnom = numpy.abs(wvno - xka)
        ra = numpy.sqrt(wvnop * wvnom)
        dpth = d[0]
        rho1 = rho[0]
        p = ra * dpth
        w, cosp, _, _, _, _, _, _, _, _, _, _ = var(
            p, q, ra, 1.0e-5, wvno, xka, xkb, dpth
        )
        dlt = cosp * e[0] - rho1 * w * e[1]
    else:
        dlt = e[0]

    return dlt


@jitted
def fast_delta(wvno, omega, d, alpha, beta, rho, llw):
    """
    Fast delta matrix.

    After Buchen and Ben-Hador (1996).

    """
    mmax = len(d)

    # Handle water layer
    if llw == 0:
        beta = beta * 1.0
        beta[0] = 1.0e-8

    # Phase velocity
    c = omega / wvno
    c2 = c * c

    # Rayleigh-wave fast delta matrix
    gam0 = beta[0] ** 2.0 / c2
    mu0 = rho[0] * beta[0] ** 2.0
    t0 = 2.0 - c2 / beta[0] ** 2.0

    X = numpy.zeros(5, dtype=numpy.complex_)
    X[0] = 2.0 * t0
    X[1] = -t0 * t0
    X[4] = -4.0
    X *= mu0 * mu0
    X, _ = normc(X, 5)

    for i in range(mmax - 1):
        scale = 0
        gam1 = beta[i + 1] ** 2.0 / c2
        eps = rho[i + 1] / rho[i]
        eta = 2.0 * (gam0 - eps * gam1)
        a = eps + eta
        ap = a - 1.0
        b = 1.0 - eta
        bp = b - 1.0
        gam0 = gam1

        r = (
            numpy.sqrt(1.0 - c2 / alpha[i] ** 2.0)
            if c < alpha[i]
            else numpy.sqrt(c2 / alpha[i] ** 2.0 - 1.0) * 1j
            if c > alpha[i]
            else 0.0
        )
        s = (
            numpy.sqrt(1.0 - c2 / beta[i] ** 2.0)
            if c < beta[i]
            else numpy.sqrt(c2 / beta[i] ** 2.0 - 1.0) * 1j
            if c > beta[i]
            else 0.0
        )

        if c < alpha[i]:
            Ca = numpy.cosh(wvno * r * d[i])
            Sa = numpy.sinh(wvno * r * d[i])
        elif c > alpha[i]:
            Ca = numpy.cos(wvno * r.imag * d[i])
            Sa = numpy.sin(wvno * r.imag * d[i]) * 1j
        else:
            Ca = 1.0
            Sa = 0.0

        if c < beta[i]:
            Cb = numpy.cosh(wvno * s * d[i])
            Sb = numpy.sinh(wvno * s * d[i])

            # Handle hyperbolic overflow
            if wvno * s.real * d[i] > 80.0:
                scale = 1
                Cb /= Ca
                Sb /= Ca

        elif c > beta[i]:
            Cb = numpy.cos(wvno * s.imag * d[i])
            Sb = numpy.sin(wvno * s.imag * d[i]) * 1j

        else:
            Cb = 1.0
            Sb = 0.0

        if i == 0 and llw == 0:
            Cb = 1.0
            Sb = 0.0

        p1 = Cb * X[1] + s * Sb * X[2]
        p2 = Cb * X[3] + s * Sb * X[4]
        p3 = Cb * X[2]
        p4 = Cb * X[4]
        if c == beta[i]:
            p3 += wvno * d[i] * X[1]
            p4 += wvno * d[i] * X[3]
        else:
            p3 += Sb * X[1] / s
            p4 += Sb * X[3] / s

        if scale == 1:
            q1 = p1 - r * p2
            q3 = p3 - r * p4
            q2 = p4
            q4 = p2
            if c != alpha[i]:
                q2 -= p3 / r
                q4 -= p1 / r
        else:
            q1 = Ca * p1 - r * Sa * p2
            q3 = Ca * p3 - r * Sa * p4
            q2 = Ca * p4
            q4 = Ca * p2
            if c == alpha[i]:
                q2 -= wvno * d[i] * p3
                q4 -= wvno * d[i] * p1
            else:
                q2 -= Sa * p3 / r
                q4 -= Sa * p1 / r

        y1 = a * q1
        y2 = ap * q2
        z1 = bp * q1
        z2 = b * q2
        if scale == 0:
            y1 += ap * X[0]
            y2 += a * X[0]
            z1 += b * X[0]
            z2 += bp * X[0]

        X[0] = bp * y1 + b * y2
        X[1] = a * y1 + ap * y2
        X[2] = eps * q3
        X[3] = eps * q4
        X[4] = bp * z1 + b * z2
        X, _ = normc(X, 5)

    r = (
        numpy.sqrt(1.0 - c2 / alpha[-1] ** 2)
        if c < alpha[-1]
        else numpy.sqrt(c2 / alpha[-1] ** 2 - 1.0) * 1j
        if c > alpha[-1]
        else 0.0
    )
    s = (
        numpy.sqrt(1.0 - c2 / beta[-1] ** 2)
        if c < beta[-1]
        else numpy.sqrt(c2 / beta[-1] ** 2 - 1.0) * 1j
        if c > beta[-1]
        else 0.0
    )

    return numpy.real(X[1] + s * X[3] - r * (X[3] + s * X[4]))


@jitted
def dltar(wvno, omega, d, a, b, rho, ifunc, llw, ca):
    """Select Rayleigh or Love wave period equation."""
    if ifunc == 1:
        return dltar1(wvno, omega, d, a, b, rho, llw)
    elif ifunc == 2:
        return dltar4(wvno, omega, d, a, b, rho, llw, ca)
    else:
        return fast_delta(wvno, omega, d, a, b, rho, llw)


@jitted
def nevill(t, c1, c2, del1, del2, d, a, b, rho, ifunc, llw, ca):
    """Hybrid method for refining root once it has been bracketted."""
    x = numpy.zeros(20, dtype=numpy.float64)
    y = numpy.zeros(20, dtype=numpy.float64)

    # Initial guess
    omega = twopi / t
    c3 = 0.5 * (c1 + c2)
    del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc, llw, ca)
    nev = 1
    nctrl = 1

    while True:
        nctrl += 1
        if nctrl >= 100:
            break

        # Make sure new estimate is inside the previous values
        # If not, perform interval halving
        if c3 < min(c1, c2) or c3 > max(c1, c2):
            nev = 0
            c3 = 0.5 * (c1 + c2)
            del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc, llw, ca)

        s13 = del1 - del3
        s32 = del3 - del2

        # Define new bounds according to the sign of the period equation
        if numpy.sign(del3) * numpy.sign(del1) < 0.0:
            c2 = c3
            del2 = del3
        else:
            c1 = c3
            del1 = del3

        # Check for convergence
        if numpy.abs(c1 - c2) <= 1.0e-6 * c1:
            break

        # If the slopes are not the same between c1, c2 and c3
        # Do not use Neville iteration
        if numpy.sign(s13) != numpy.sign(s32):
            nev = 0

        # If the period equation differs by more than a factor of 10
        # Use interval halving to avoid poor behavior of polynomial fit
        ss1 = numpy.abs(del1)
        s1 = 0.01 * ss1
        ss2 = numpy.abs(del2)
        s2 = 0.01 * ss2
        if s1 > ss2 or s2 > ss1 or nev == 0:
            c3 = 0.5 * (c1 + c2)
            del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc, llw, ca)
            nev = 1
            m = 1
        else:
            if nev == 2:
                x[m - 1] = c3
                y[m - 1] = del3
            else:
                x[0] = c1
                y[0] = del1
                x[1] = c2
                y[1] = del2
                m = 1

            # Perform Neville iteration
            flag = 1
            for kk in range(m):
                j = m - kk
                denom = y[m] - y[j]
                if numpy.abs(denom) < 1.0e-10 * numpy.abs(y[m]):
                    flag = 0
                    break
                else:
                    x[j - 1] = (-y[j - 1] * x[j] + y[m] * x[j - 1]) / denom

            if flag:
                c3 = x[0]
                del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc, llw, ca)
                nev = 2
                m += 1
                m = min(m, 10)
            else:
                c3 = 0.5 * (c1 + c2)
                del3 = dltar(omega / c3, omega, d, a, b, rho, ifunc, llw, ca)
                nev = 1
                m = 1

    return c3


@jitted
def getsol(t1, c1, clow, dc, cm, betmx, ifirst, del1st, d, a, b, rho, ifunc, llw, ca):
    """Bracket dispersion curve and then refine it."""
    # Bracket solution
    omega = twopi / t1
    del1 = dltar(omega / c1, omega, d, a, b, rho, ifunc, llw, ca)
    del1st = del1 if ifirst else del1st
    idir = -1.0 if not ifirst and numpy.sign(del1st) * numpy.sign(del1) < 0.0 else 1.0

    # idir indicates the direction of the search for the true phase velocity from the initial estimate
    while True:
        c2 = c1 + idir * dc

        if c2 <= clow:
            idir = 1.0
            c1 = clow
        else:
            omega = twopi / t1
            del2 = dltar(omega / c2, omega, d, a, b, rho, ifunc, llw, ca)

            if numpy.sign(del1) != numpy.sign(del2):
                c1 = nevill(t1, c1, c2, del1, del2, d, a, b, rho, ifunc, llw, ca)
                iret = c1 > betmx
                break

            c1 = c2
            del1 = del2

            iret = c1 < cm or c1 >= betmx + dc
            if iret:
                break

    return c1, del1st, iret


@jitted
def gtsolh(a, b):
    """Starting solution."""
    c = 0.95 * b

    for _ in range(5):
        gamma = b / a
        kappa = c / b
        k2 = kappa ** 2
        gk2 = (gamma * kappa) ** 2
        fac1 = numpy.sqrt(1.0 - gk2)
        fac2 = numpy.sqrt(1.0 - k2)
        fr = (2.0 - k2) ** 2 - 4.0 * fac1 * fac2
        frp = -4.0 * (2.0 - k2) * kappa
        frp += 4.0 * fac2 * gamma * gamma * kappa / fac1
        frp += 4.0 * fac1 * kappa / fac2
        frp /= b
        c -= fr / frp

    return c


@jitted
def getc(t, d, a, b, rho, mode, ifunc, dc):
    """Get phase velocity dispersion curve."""
    # Initialize arrays
    kmax = len(t)
    c = numpy.zeros(kmax, dtype=numpy.float64)
    cg = numpy.zeros(kmax, dtype=numpy.float64)

    # Preallocate Dunkin's matrix
    ca = numpy.empty((5, 5), dtype=numpy.float64)

    # Check for water layer
    llw = 0 if b[0] <= 0.0 else -1

    # Find the extremal velocities to assist in starting search
    betmx = -1.0e20
    betmn = 1.0e20
    mmax = len(d)
    for i in range(mmax):
        if b[i] > 0.01 and b[i] < betmn:
            betmn = b[i]
            jmn = i
            jsol = False
        elif b[i] < 0.01 and a[i] < betmn:
            betmn = a[i]
            jmn = i
            jsol = True
        betmx = max(betmx, b[i])

    # Solid layer solve halfspace period equation
    cc = betmn if jsol else gtsolh(a[jmn], b[jmn])

    # Back off a bit to get a starting value at a lower phase velocity
    cc *= 0.9
    c1 = cc
    cm = cc

    one = 1.0e-2
    onea = 1.5
    for iq in range(mode + 1):
        ibeg = 0
        iend = kmax

        del1st = 0.0
        for k in range(ibeg, iend):
            # Get initial phase velocity estimate to begin search
            ifirst = k == ibeg
            if ifirst and iq == 0:
                clow = cc
                c1 = cc
            elif ifirst and iq > 0:
                clow = c1
                c1 = c[ibeg] + one * dc
            elif not ifirst and iq > 0:
                clow = c[k] + one * dc
                c1 = max(c[k - 1], clow)
            elif not ifirst and iq == 0:
                clow = cm
                c1 = c[k - 1] - onea * dc

            # Bracket root and refine it
            c1, del1st, iret = getsol(
                t[k],
                c1,
                clow,
                dc,
                cm,
                betmx,
                ifirst,
                del1st,
                d,
                a,
                b,
                rho,
                ifunc,
                llw,
                ca,
            )

            if iret:
                if iq > 0:
                    for i in range(k, kmax):
                        cg[i] = 0.0

                    if iq == mode:
                        return cg
                    else:
                        c1 = 0.0
                        break
                else:
                    raise DispersionError("failed to find root for fundamental mode")

            c[k] = c1
            cg[k] = c[k]
            c1 = 0.0

    return cg


@jitted
def surf96(t, d, a, b, rho, mode=0, itype=0, ifunc=2, dc=0.005, dt=0.025):
    """
    Get phase or group velocity dispersion curve.

    Parameters
    ----------
    t : array_like
        Periods (in s).
    d : array_like
        Layer thickness (in km).
    a : array_like
        Layer P-wave velocity (in km/s).
    b : array_like
        Layer S-wave velocity (in km/s).
    rho : array_like
        Layer density (in g/cm3).
    mode : int, optional, default 0
        Mode number (0 if fundamental).
    itype : int, optional, default 0
        Velocity type:
         - 0: phase velocity,
         - 1: group velocity.
    ifunc : int, optional, default 2
        Select wave type and algorithm for period equation:
         - 1: Love-wave (Thomson-Haskell method),
         - 2: Rayleigh-wave (Dunkin's matrix),
         - 3: Rayleigh-wave (fast delta matrix).
    dc : scalar, optional, default 0.005
        Phase velocity increment for root finding.
    dt : scalar, optional, default 0.025
        Frequency increment (%) for calculating group velocity.

    Returns
    -------
    array_like
        Phase or group dispersion velocity.

    """
    nt = len(t)
    t1 = numpy.empty(nt, dtype=numpy.float64)

    if itype == 1:
        fac = 1.0 + dt
        for i in range(nt):
            t1[i] = t[i] / fac
    else:
        for i in range(nt):
            t1[i] = t[i]

    c1 = getc(t1, d, a, b, rho, mode, ifunc, dc)

    if itype == 1:
        t2 = numpy.empty(nt, dtype=numpy.float64)
        fac = 1.0 - dt
        for i in range(nt):
            t2[i] = t[i] / fac

        c2 = getc(t2, d, a, b, rho, mode, ifunc, dc)

        for i in range(nt):
            if c2[i] > 0.0:
                t1i = 1.0 / t1[i]
                t2i = 1.0 / t2[i]
                c1[i] = (t1i - t2i) / (t1i / c1[i] - t2i / c2[i])
            else:
                c1[i] = 0.0

    return c1



import os

import helpers
import matplotlib.pyplot as plt
import numpy
import pytest

import disba

if not os.environ.get("DISPLAY", ""):
    plt.switch_backend("Agg")


@pytest.mark.parametrize(
    "mode, wave, algorithm",
    [
        (0, "rayleigh", "dunkin"),
        (0, "rayleigh", "fast-delta"),
        (1, "rayleigh", "dunkin"),
        (1, "rayleigh", "fast-delta"),
        (0, "love", "dunkin"),
        (1, "love", "dunkin"),
    ],
)
def test_resample(mode, wave, algorithm):
    velocity_model = helpers.velocity_model(5)
    t = numpy.logspace(0.0, 1.0, 20)

    pd = disba.PhaseDispersion(*velocity_model, algorithm=algorithm)
    cref = pd(t, mode, wave)

    pd.resample(0.1)
    cp = pd(t, mode, wave)

    assert numpy.allclose(cref.velocity.sum(), cp.velocity.sum(), atol=0.1)


def test_depthplot(monkeypatch):
    velocity_model = helpers.velocity_model(5)

    monkeypatch.setattr(plt, "show", lambda: None)
    disba.depthplot(velocity_model[0], velocity_model[2])



from abc import ABC

import numpy

from ._helpers import resample

algorithms = {"dunkin", "fast-delta"}


def is_arraylike(arr, size):
    """Check input array."""
    return isinstance(arr, (list, tuple, numpy.ndarray)) and numpy.size(arr) == size


class Base(ABC):
    def __init__(self, thickness, velocity_p, velocity_s, density, algorithm, dc):
        """Base class."""
        mmax = len(thickness)
        if not is_arraylike(thickness, mmax):
            raise TypeError()
        if not is_arraylike(velocity_p, mmax):
            raise TypeError()
        if not is_arraylike(velocity_s, mmax):
            raise TypeError()
        if not is_arraylike(density, mmax):
            raise TypeError()
        if algorithm not in algorithms:
            raise ValueError()
        if not isinstance(dc, float):
            raise TypeError()

        self._thickness = numpy.asarray(thickness)
        self._velocity_p = numpy.asarray(velocity_p)
        self._velocity_s = numpy.asarray(velocity_s)
        self._density = numpy.asarray(density)
        self._algorithm = algorithm
        self._dc = dc

    def resample(self, dz):
        """
        Resample velocity model.

        Parameters
        ----------
        dz : scalar
            Maximum layer thickness (in km).

        """
        if dz <= 0.0:
            raise ValueError()

        d, a, b, rho = resample(
            self._thickness, self._velocity_p, self._velocity_s, self._density, dz,
        )
        self._thickness = d
        self._velocity_p = a
        self._velocity_s = b
        self._density = rho

    @property
    def thickness(self):
        """Return layer thickness (in km)."""
        return self._thickness

    @property
    def velocity_p(self):
        """Return layer P-wave velocity (in km/s)."""
        return self._velocity_p

    @property
    def velocity_s(self):
        """Return layer S-wave velocity (in km/s)."""
        return self._velocity_s

    @property
    def density(self):
        """Return layer density (in g/cm3)."""
        return self._density

    @property
    def algorithm(self):
        """Return algorithm to use for computation of Rayleigh-wave dispersion."""
        return self._algorithm

    @property
    def dc(self):
        """Return phase velocity increment for root finding."""
        return self._dc


class BaseDispersion(Base):
    def __init__(self, thickness, velocity_p, velocity_s, density, algorithm, dc):
        """Base class for dispersion."""
        super().__init__(thickness, velocity_p, velocity_s, density, algorithm, dc)


class BaseSensitivity(Base):
    def __init__(self, thickness, velocity_p, velocity_s, density, algorithm, dc, dp):
        """Base class for sensitivity kernel."""
        if not isinstance(dp, float):
            raise TypeError()

        super().__init__(thickness, velocity_p, velocity_s, density, algorithm, dc)

        self._dp = dp

    @property
    def dp(self):
        """Return parameter increment (%) for numerical partial derivatives."""
        return self._dp



from numba import jit

ifunc = {
    "dunkin": {"love": 1, "rayleigh": 2},
    "fast-delta": {"love": 1, "rayleigh": 3},
}

ipar = {
    "thickness": 0,
    "velocity_p": 1,
    "velocity_s": 2,
    "density": 3,
}


def jitted(*args, **kwargs):
    """Custom :func:`jit` with default options."""
    kwargs.update(
        {
            "nopython": True,
            "nogil": True,
            "fastmath": True,
            # "boundscheck": False,
            "cache": True,
        }
    )
    return jit(*args, **kwargs)



import numpy
import pytest

from disba import DispersionError, PhaseDispersion


def test_exception():
    velocity_model = numpy.array(
        [[0.5, 1.0, 0.5, 2.00], [0.3, 2.0, 1.0, 2.00], [1.0, 1.0, 0.5, 2.00]]
    )
    f = numpy.linspace(0.1, 10.0, 60)
    t = 1.0 / f[::-1]

    with pytest.raises(DispersionError):
        pd = PhaseDispersion(*velocity_model.T, algorithm="dunkin")
        pd(t, mode=0, wave="rayleigh")



from collections import namedtuple

import numpy

from ._base import BaseSensitivity
from ._common import ifunc, ipar
from ._cps import srfker96, swegn96

__all__ = [
    "SensitivityKernel",
    "PhaseSensitivity",
    "GroupSensitivity",
    "EllipticitySensitivity",
]


SensitivityKernel = namedtuple(
    "SensitivityKernel",
    ("depth", "kernel", "period", "velocity", "mode", "wave", "type", "parameter"),
)


class PhaseSensitivity(BaseSensitivity):
    def __init__(
        self,
        thickness,
        velocity_p,
        velocity_s,
        density,
        algorithm="dunkin",
        dc=0.005,
        dp=0.025,
    ):
        """
        Phase velocity sensitivity kernel class.

        Parameters
        ----------
        thickness : array_like
            Layer thickness (in km).
        velocity_p : array_like
            Layer P-wave velocity (in km/s).
        velocity_s : array_like
            Layer S-wave velocity (in km/s).
        density : array_like
            Layer density (in g/cm3).
        algorithm : str {'dunkin', 'fast-delta'}, optional, default 'dunkin'
            Algorithm to use for computation of Rayleigh-wave dispersion:
             - 'dunkin': Dunkin's matrix (adapted from surf96),
             - 'fast-delta': fast delta matrix (after Buchen and Ben-Hador, 1996).
        dc : scalar, optional, default 0.005
            Phase velocity increment for root finding.
        dp : scalar, optional, default 0.025
            Parameter increment (%) for numerical partial derivatives.

        """
        super().__init__(thickness, velocity_p, velocity_s, density, algorithm, dc, dp)

    def __call__(self, t, mode=0, wave="rayleigh", parameter="velocity_s"):
        """
        Calculate phase velocity sensitivity kernel for a given period and parameter.

        Parameters
        ----------
        t : scalar
            Period (in s).
        mode : int, optional, default 0
            Mode number (0 if fundamental).
        wave : str {'love', 'rayleigh'}, optional, default 'rayleigh'
            Wave type.
        parameter : str {'thickness', 'velocity_p', 'velocity_s', 'density'}, optional, default 'velocity_s'
            Parameter with respect to which sensitivity kernel is calculated.

        Returns
        -------
        :class:`disba.SensitivityKernel`
            Sensitivity kernel as a namedtuple (depth, kernel, period, velocity, mode, wave, type, parameter).

        """
        c1, kernel = srfker96(
            t,
            self._thickness,
            self._velocity_p,
            self._velocity_s,
            self._density,
            mode=mode,
            itype=0,
            ifunc=ifunc[self._algorithm][wave],
            ipar=ipar[parameter],
            dc=self._dc,
            dp=self._dp,
        )

        return SensitivityKernel(
            numpy.insert(self._thickness.cumsum()[:-1], 0, 0.0),
            kernel,
            t,
            c1,
            mode,
            wave,
            "phase",
            parameter,
        )


class GroupSensitivity(BaseSensitivity):
    def __init__(
        self,
        thickness,
        velocity_p,
        velocity_s,
        density,
        algorithm="dunkin",
        dc=0.005,
        dt=0.025,
        dp=0.025,
    ):
        """
        Phase velocity sensitivity kernel class.

        Parameters
        ----------
        thickness : array_like
            Layer thickness (in km).
        velocity_p : array_like
            Layer P-wave velocity (in km/s).
        velocity_s : array_like
            Layer S-wave velocity (in km/s).
        density : array_like
            Layer density (in g/cm3).
        algorithm : str {'dunkin', 'fast-delta'}, optional, default 'dunkin'
            Algorithm to use for computation of Rayleigh-wave dispersion:
             - 'dunkin': Dunkin's matrix (adapted from surf96),
             - 'fast-delta': fast delta matrix (after Buchen and Ben-Hador, 1996).
        dc : scalar, optional, default 0.005
            Phase velocity increment for root finding.
        dt : scalar, optional, default 0.025
            Frequency increment (%) for calculating group velocity.
        dp : scalar, optional, default 0.025
            Parameter increment (%) for numerical partial derivatives.

        """
        if not isinstance(dt, float):
            raise TypeError()

        super().__init__(thickness, velocity_p, velocity_s, density, algorithm, dc, dp)

        self._dt = dt

    def __call__(self, t, mode=0, wave="rayleigh", parameter="velocity_s"):
        """
        Calculate group velocity sensitivity kernel for a given period and parameter.

        Parameters
        ----------
        t : scalar
            Period (in s).
        mode : int, optional, default 0
            Mode number (0 if fundamental).
        wave : str {'love', 'rayleigh'}, optional, default 'rayleigh'
            Wave type.
        parameter : str {'thickness', 'velocity_p', 'velocity_s', 'density'}, optional, default 'velocity_s'
            Parameter with respect to which sensitivity kernel is calculated.

        Returns
        -------
        :class:`disba.SensitivityKernel`
            Sensitivity kernel as a namedtuple (depth, kernel, period, velocity, mode, wave, type, parameter).

        """
        c1, kernel = srfker96(
            t,
            self._thickness,
            self._velocity_p,
            self._velocity_s,
            self._density,
            mode=mode,
            itype=1,
            ifunc=ifunc[self._algorithm][wave],
            ipar=ipar[parameter],
            dc=self._dc,
            dt=self._dt,
            dp=self._dp,
        )

        return SensitivityKernel(
            numpy.insert(self._thickness.cumsum()[:-1], 0, 0.0),
            kernel,
            t,
            c1,
            mode,
            wave,
            "group",
            parameter,
        )

    @property
    def dt(self):
        """Return frequency increment (%) for calculating group velocity."""
        return self._dt


class EllipticitySensitivity(BaseSensitivity):
    def __init__(
        self,
        thickness,
        velocity_p,
        velocity_s,
        density,
        algorithm="dunkin",
        dc=0.005,
        dp=0.025,
    ):
        """
        Rayleigh-wave ellipticity sensitivity kernel class.

        Parameters
        ----------
        thickness : array_like
            Layer thickness (in km).
        velocity_p : array_like
            Layer P-wave velocity (in km/s).
        velocity_s : array_like
            Layer S-wave velocity (in km/s).
        density : array_like
            Layer density (in g/cm3).
        algorithm : str {'dunkin', 'fast-delta'}, optional, default 'dunkin'
            Algorithm to use for computation of Rayleigh-wave dispersion:
             - 'dunkin': Dunkin's matrix (adapted from surf96),
             - 'fast-delta': fast delta matrix (after Buchen and Ben-Hador, 1996).
        dc : scalar, optional, default 0.005
            Phase velocity increment for root finding.
        dp : scalar, optional, default 0.025
            Parameter increment (%) for numerical partial derivatives.

        """
        super().__init__(thickness, velocity_p, velocity_s, density, algorithm, dc, dp)

    def __call__(self, t, mode=0, parameter="velocity_s"):
        """
        Calculate Rayleigh-wave ellipticity sensitivity kernel for a given period and
        parameter.

        Parameters
        ----------
        t : scalar
            Period (in s).
        mode : int, optional, default 0
            Mode number (0 if fundamental).
        parameter : str {'thickness', 'velocity_p', 'velocity_s', 'density'}, optional, default 'velocity_s'
            Parameter with respect to which sensitivity kernel is calculated.

        Returns
        -------
        :class:`disba.SensitivityKernel`
            Sensitivity kernel as a namedtuple (depth, kernel, period, velocity, mode, wave, type, parameter).

        """
        # Reference ellipticity
        ell1 = self._ellipticity(t, mode)

        # Initialize kernel
        mmax = len(self._thickness)
        kernel = numpy.empty(mmax)

        # Loop over layers
        fac = 1.0 + self._dp
        par = getattr(self, parameter)
        for i in range(mmax):
            tmp = par[i]
            par[i] /= fac
            ell2 = self._ellipticity(t, mode)
            kernel[i] = (ell2 - ell1) / (par[i] - tmp)
            par[i] *= fac

        return SensitivityKernel(
            numpy.insert(self._thickness.cumsum()[:-1], 0, 0.0),
            kernel,
            t,
            None,
            mode,
            "rayleigh",
            "ellipticity",
            parameter,
        )

    def _ellipticity(self, t, mode):
        """Compute Rayleigh-wave ellipticity for input period and mode."""
        eig = swegn96(
            t,
            self._thickness,
            self._velocity_p,
            self._velocity_s,
            self._density,
            mode,
            ifunc[self._algorithm]["rayleigh"],
            self._dc,
        )[:, :2]

        return eig[0, 0] / eig[0, 1]



import helpers
import numpy
import pytest

import disba


@pytest.mark.parametrize(
    "wave, parameter, kref, atol",
    [
        ("rayleigh", "thickness", -0.130, 1.0e-3),
        ("rayleigh", "velocity_p", 0.173, 1.0e-3),
        ("rayleigh", "velocity_s", 0.722, 1.0e-3),
        ("rayleigh", "density", 0.000245, 1.0e-6),
        ("love", "thickness", -0.197, 1.0e-3),
        ("love", "velocity_p", 0.0, 1.0e-3),
        ("love", "velocity_s", 1.194, 1.0e-3),
        ("love", "density", 0.000368, 1.0e-6),
    ],
)
def test_phase(wave, parameter, kref, atol):
    velocity_model = helpers.velocity_model(5)

    ps = disba.PhaseSensitivity(*velocity_model, dp=0.005, algorithm="dunkin")
    kp = ps(10.0, 0, wave, parameter)

    assert numpy.allclose(kref, kp.kernel.sum(), atol=atol)


@pytest.mark.parametrize(
    "wave, parameter, kref, atol",
    [
        ("rayleigh", "thickness", -0.252, 1.0e-3),
        ("rayleigh", "velocity_p", 0.254, 1.0e-3),
        ("rayleigh", "velocity_s", 0.804, 1.0e-3),
        ("rayleigh", "density", 0.0207, 1.0e-4),
        ("love", "thickness", -0.195, 1.0e-3),
        ("love", "velocity_p", 0.0, 1.0e-3),
        ("love", "velocity_s", 1.332, 1.0e-3),
        ("love", "density", 0.0479, 1.0e-4),
    ],
)
def test_group(wave, parameter, kref, atol):
    velocity_model = helpers.velocity_model(5)

    gs = disba.GroupSensitivity(*velocity_model, dt=0.005, dp=0.005, algorithm="dunkin")
    kg = gs(10.0, 0, wave, parameter)

    assert numpy.allclose(kref, kg.kernel.sum(), atol=atol)


@pytest.mark.parametrize(
    "parameter, kref, atol",
    [
        ("thickness", -0.293, 1.0e-3),
        ("velocity_p", -0.450, 1.0e-3),
        ("velocity_s", 0.731, 1.0e-3),
        ("density", -0.000230, 1.0e-6),
    ],
)
def test_ellipticity(parameter, kref, atol):
    velocity_model = helpers.velocity_model(5)

    es = disba.EllipticitySensitivity(*velocity_model, dp=0.005, algorithm="dunkin")
    ke = es(10.0, 0, parameter)

    assert numpy.allclose(kref, ke.kernel.sum(), atol=atol)



from collections import namedtuple

import numpy

from ._base import Base
from ._common import ifunc
from ._cps import swegn96

__all__ = [
    "LoveEigen",
    "RayleighEigen",
    "EigenFunction",
]


LoveEigen = namedtuple("LoveEigen", ("depth", "uu", "tt", "period", "mode"))

RayleighEigen = namedtuple(
    "RayleighEigen", ("depth", "ur", "uz", "tz", "tr", "period", "mode")
)


class EigenFunction(Base):
    def __init__(
        self, thickness, velocity_p, velocity_s, density, algorithm="dunkin", dc=0.005,
    ):
        """
        Eigenfunction class.

        Parameters
        ----------
        thickness : array_like
            Layer thickness (in km).
        velocity_p : array_like
            Layer P-wave velocity (in km/s).
        velocity_s : array_like
            Layer S-wave velocity (in km/s).
        density : array_like
            Layer density (in g/cm3).
        algorithm : str {'dunkin', 'fast-delta'}, optional, default 'dunkin'
            Algorithm to use for computation of Rayleigh-wave dispersion:
             - 'dunkin': Dunkin's matrix (adapted from surf96),
             - 'fast-delta': fast delta matrix (after Buchen and Ben-Hador, 1996).
        dc : scalar, optional, default 0.005
            Phase velocity increment for root finding.

        """
        super().__init__(thickness, velocity_p, velocity_s, density, algorithm, dc)

    def __call__(self, t, mode=0, wave="rayleigh"):
        """
        Compute eigenfunctions for a given period and mode.

        Parameters
        ----------
        t : scalar
            Period (in s).
        mode : int, optional, default 0
            Mode number (0 if fundamental).
        wave : str {'love', 'rayleigh'}, optional, default 'rayleigh'
            Wave type.

        Returns
        -------
        :class:`disba.RayleighEigen` or :class:`disba.LoveEigen`
            Eigenfunction as a namedtuple:

             - If ``wave == 'love'``, (depth, uu, tt, period, mode),
             - If ``wave == 'rayleigh'``, (depth, ur, uz, tz, tr, period, mode).

        """
        if numpy.ndim(t) > 0:
            raise ValueError("Period t must be scalar.")

        egn = swegn96(
            t,
            self._thickness,
            self._velocity_p,
            self._velocity_s,
            self._density,
            mode,
            ifunc[self._algorithm][wave],
            self._dc,
        )

        depth = self._thickness.cumsum() - self._thickness[0]
        if wave == "love":
            uu, tt = egn.T
            wegn = LoveEigen(depth, uu, tt, t, mode)
        elif wave == "rayleigh":
            ur, uz, tz, tr = egn.T
            wegn = RayleighEigen(depth, ur, uz, tz, tr, t, mode)

        return wegn



from collections import namedtuple

from ._base import BaseDispersion
from ._common import ifunc
from ._cps import surf96
from ._helpers import is_sorted

__all__ = [
    "DispersionCurve",
    "PhaseDispersion",
    "GroupDispersion",
]


DispersionCurve = namedtuple(
    "DispersionCurve", ("period", "velocity", "mode", "wave", "type")
)


class PhaseDispersion(BaseDispersion):
    def __init__(
        self, thickness, velocity_p, velocity_s, density, algorithm="dunkin", dc=0.005,
    ):
        """
        Phase velocity dispersion class.

        Parameters
        ----------
        thickness : array_like
            Layer thickness (in km).
        velocity_p : array_like
            Layer P-wave velocity (in km/s).
        velocity_s : array_like
            Layer S-wave velocity (in km/s).
        density : array_like
            Layer density (in g/cm3).
        algorithm : str {'dunkin', 'fast-delta'}, optional, default 'dunkin'
            Algorithm to use for computation of Rayleigh-wave dispersion:
             - 'dunkin': Dunkin's matrix (adapted from surf96),
             - 'fast-delta': fast delta matrix (after Buchen and Ben-Hador, 1996).
        dc : scalar, optional, default 0.005
            Phase velocity increment for root finding.

        """
        super().__init__(thickness, velocity_p, velocity_s, density, algorithm, dc)

    def __call__(self, t, mode=0, wave="rayleigh"):
        """
        Calculate phase velocities for input period axis.

        Parameters
        ----------
        t : array_like
            Periods (in s).
        mode : int, optional, default 0
            Mode number (0 if fundamental).
        wave : str {'love', 'rayleigh'}, optional, default 'rayleigh'
            Wave type.

        Returns
        -------
        :class:`disba.DispersionCurve`
            Dispersion curve as a namedtuple (period, velocity, mode, wave, type).

        """
        if not is_sorted(t):
            raise ValueError("period axis must be sorted")

        c = surf96(
            t,
            self._thickness,
            self._velocity_p,
            self._velocity_s,
            self._density,
            mode,
            0,
            ifunc[self._algorithm][wave],
            self._dc,
        )

        idx = c > 0.0
        t = t[idx]
        c = c[idx]

        return DispersionCurve(t, c, mode, wave, "phase")


class GroupDispersion(BaseDispersion):
    def __init__(
        self,
        thickness,
        velocity_p,
        velocity_s,
        density,
        algorithm="dunkin",
        dc=0.005,
        dt=0.025,
    ):
        """
        Group velocity dispersion class.

        Parameters
        ----------
        thickness : array_like
            Layer thickness (in km).
        velocity_p : array_like
            Layer P-wave velocity (in km/s).
        velocity_s : array_like
            Layer S-wave velocity (in km/s).
        density : array_like
            Layer density (in g/cm3).
        algorithm : str {'dunkin', 'fast-delta'}, optional, default 'dunkin'
            Algorithm to use for computation of Rayleigh-wave dispersion:
             - 'dunkin': Dunkin's matrix (adapted from surf96),
             - 'fast-delta': fast delta matrix (after Buchen and Ben-Hador, 1996).
        dc : scalar, optional, default 0.005
            Phase velocity increment for root finding.
        dt : scalar, optional, default 0.025
            Frequency increment (%) for calculating group velocity.

        """
        if not isinstance(dt, float):
            raise TypeError()

        super().__init__(thickness, velocity_p, velocity_s, density, algorithm, dc)

        self._dt = dt

    def __call__(self, t, mode=0, wave="rayleigh"):
        """
        Calculate group velocities for input period axis.

        Parameters
        ----------
        t : array_like
            Periods (in s).
        mode : int, optional, default 0
            Mode number (0 if fundamental).
        wave : str {'love', 'rayleigh'}, optional, default 'rayleigh'
            Wave type.

        Returns
        -------
        :class:`disba.DispersionCurve`
            Dispersion curve as a namedtuple (period, velocity, mode, wave, type).

        """
        if not is_sorted(t):
            raise ValueError("period axis must be sorted")

        c = surf96(
            t,
            self._thickness,
            self._velocity_p,
            self._velocity_s,
            self._density,
            mode,
            1,
            ifunc[self._algorithm][wave],
            self._dc,
            self._dt,
        )

        idx = c > 0.0
        t = t[idx]
        c = c[idx]

        return DispersionCurve(t, c, mode, wave, "group")

    @property
    def dt(self):
        """Return frequency increment (%) for calculating group velocity."""
        return self._dt



"""
Numba implementation of the Fortran programs sregn96 and slegn96.

This module is not a one-to-one translation from Fortran to Python.
The code has been adapted and optimized for Numba.

Only the computation of eigenfunctions is implemented.

..

    COMPUTER PROGRAMS IN SEISMOLOGY
    VOLUME III

    COPYRIGHT 1986, 1991
    D. R. Russell, R. B. Herrmann
    Department of Earth and Atmospheric Sciences
    Saint Louis University
    221 North Grand Boulevard
    St. Louis, Missouri 63103
    U. S. A.

"""

import numpy

from .._common import jitted
from .._exception import DispersionError
from ._common import normc
from ._surf96 import surf96

__all__ = [
    "swegn96",
]


@jitted
def evalg(m, d, a, b, rho, wvno, om):
    """Layered half space problem for Rayleigh-wave."""
    gbr = numpy.zeros(5, dtype=numpy.complex_)
    wvno2 = wvno * wvno
    om2 = om * om

    # Set up halfspace conditions
    xka = om / a[m]
    xkb = om / b[m] if b[m] > 0.01 else 0.0
    ra = numpy.sqrt(wvno2 - xka * xka + 0.0 * 1j)
    rb = numpy.sqrt(wvno2 - xkb * xkb + 0.0 * 1j)
    gam = b[m] * wvno / om
    gam = 2.0 * gam * gam
    gamm1 = gam - (1.0 + 0.0 * 1j)

    # Half space
    if b[m] > 0.01:
        gbr[0] = (
            rho[m] * rho[m] * om2 * om2 * (-gam * gam * ra * rb + wvno2 * gamm1 * gamm1)
        )
        gbr[1] = -rho[m] * wvno2 * ra * om2
        gbr[2] = -rho[m] * (-gam * ra * rb + wvno2 * gamm1) * om2 * wvno
        gbr[3] = rho[m] * wvno2 * rb * om2
        gbr[4] = wvno2 * (wvno2 - ra * rb)

        fac = 0.25 / (-rho[m] * rho[m] * om2 * om2 * wvno2 * ra * rb)
        for i in range(5):
            gbr[i] *= fac

    else:
        # All fluid
        if numpy.all(b < 0.01):
            gbr[0] = 0.5 / ra
            gbr[1] = (0.5 + 0.0 * 1j) / (-rho[m] * om2)
            gbr[2] = 0.0 + 0.0 * 1j
            gbr[3] = 0.0 + 0.0 * 1j
            gbr[4] = 0.0 + 0.0 * 1j
        else:
            gbr[0] = 0.0 + 0.0 * 1j
            gbr[1] = 0.0 + 0.0 * 1j
            gbr[2] = 0.0 + 0.0 * 1j
            gbr[3] = 0.5 * rho[m] * om2 / ra
            gbr[4] = -0.5 + 0.0 * 1j

    return numpy.real(gbr)


@jitted
def varl(m, omega, wvno, dpth, b, rho):
    """Find variables cosQ, sinQ for Love-wave."""
    # Define the horizontal wavenumber for the S-wave
    xkb = omega / b[m]

    # Define the vertical wavenumber for the given wvno
    wvnop = wvno + xkb
    wvnom = numpy.abs(wvno - xkb)
    fac = wvnop * wvnom
    rb = numpy.sqrt(wvnop * wvnom)
    q = rb * dpth
    mu = rho[m] * b[m] * b[m]

    # Examine S-wave eigenfunctions
    # Checking whether c > vs, c = vs or c < vs
    eexl = 0.0
    if wvno < xkb:
        sinq = numpy.sin(q)
        y = sinq / rb
        z = -rb * sinq
        cosq = numpy.cos(q)
    elif wvno == xkb:
        cosq = 1.0
        y = dpth
        z = 0.0
    else:
        eexl = q
        fac = numpy.exp(-2.0 * q) if q < 18.0 else 0.0
        cosq = (1.0 + fac) * 0.5
        sinq = (1.0 - fac) * 0.5
        y = sinq / rb
        z = rb * sinq

    return rb, mu, cosq, sinq, y, z, eexl


@jitted
def varsv(p, q, rp, rsv, d, iwat):
    """Find variables cosP, cosQ, sinP, sinQ for Rayleigh-wave."""
    pr = numpy.real(p)
    pi = numpy.imag(p)
    qr = numpy.real(q)
    qi = numpy.imag(q)

    pex = pr
    svex = 0.0

    epp = 0.5 * (numpy.cos(pi) + numpy.sin(pi) * 1j)
    epm = numpy.conj(epp)
    pfac = numpy.exp(-2.0 * pr) if pr < 30.0 else 0.0
    cosp = numpy.real(epp + pfac * epm)
    sinp = epp - pfac * epm
    rsinp = numpy.real(rp * sinp)
    sinpr = d if numpy.abs(pr) < 1.0e-5 and numpy.abs(rp) < 1.0e-5 else sinp / rp

    # Fluid layer
    if iwat == 1:
        cosq = 1.0
        rsinq = 0.0
        sinqr = 0.0

    # Elastic layer
    else:
        svex = qr
        eqp = 0.5 * (numpy.cos(qi) + numpy.sin(qi) * 1j)
        eqm = numpy.conj(eqp)
        svfac = numpy.exp(-2.0 * qr) if qr < 30.0 else 0.0
        cosq = numpy.real(eqp + svfac * eqm)
        sinq = eqp - svfac * eqm
        rsinq = numpy.real(rsv * sinq)
        sinqr = d if numpy.abs(qr) < 1.0e-5 and numpy.abs(rsv) < 1.0e-5 else sinq / rsv

    return cosp, cosq, rsinp, rsinq, sinpr, sinqr, pex, svex


@jitted
def hskl(m, b, mu, cosq, y, z):
    """Thomson-Haskell's matrix for Love-wave."""
    hl = numpy.zeros((2, 2), dtype=numpy.float64)

    if b[m] > 0.01:
        hl[0, 0] = cosq
        hl[0, 1] = y / mu
        hl[1, 0] = z * mu
        hl[1, 1] = cosq
    else:
        hl[0, 0] = 1.0
        hl[0, 1] = 0.0
        hl[1, 0] = 0.0
        hl[1, 1] = 1.0

    return hl


@jitted
def hska(
    omega, wvno, b, rho, cosp, rsinp, sinpr, tcossv, trsinsv, tsinsvr, pex, svex, iwat
):
    """Thomson-Haskell's matrix for Rayleigh-wave."""
    aa = numpy.zeros((4, 4), dtype=numpy.complex_)
    wvno2 = wvno * wvno
    om2 = omega * omega

    if iwat == 1:
        # Fluid layer
        dfac = numpy.exp(-pex) if pex < 35.0 else 0.0
        aa[0, 0] = dfac
        aa[3, 3] = dfac
        aa[1, 1] = cosp
        aa[2, 2] = cosp
        aa[1, 2] = -rsinp / rho / om2
        aa[2, 1] = -rho * om2 * sinpr
    else:
        # Elastic layer
        dfac = numpy.exp(svex - pex) if pex - svex < 70.0 else 0.0
        cossv = dfac * tcossv
        rsinsv = dfac * trsinsv
        sinsvr = dfac * tsinsvr
        gam = 2.0 * b * b * wvno2 / om2
        gamm1 = gam - 1.0

        aa[0, 0] = cossv + gam * (cosp - cossv)
        aa[0, 1] = -wvno * gamm1 * sinpr + gam * rsinsv / wvno
        aa[0, 2] = -wvno * (cosp - cossv) / rho / om2
        aa[0, 3] = (wvno2 * sinpr - rsinsv) / rho / om2
        aa[1, 0] = gam * rsinp / wvno - wvno * gamm1 * sinsvr
        aa[1, 1] = cosp - gam * (cosp - cossv)
        aa[1, 2] = (-rsinp + wvno2 * sinsvr) / rho / om2
        aa[1, 3] = -aa[0, 2]
        aa[2, 0] = rho * om2 * gam * gamm1 * (cosp - cossv) / wvno
        aa[2, 1] = rho * om2 * (-gamm1 * gamm1 * sinpr + gam * gam * rsinsv / wvno2)
        aa[2, 2] = aa[1, 1]
        aa[2, 3] = -aa[0, 1]
        aa[3, 0] = rho * om2 * (gam * gam * rsinp / wvno2 - gamm1 * gamm1 * sinsvr)
        aa[3, 1] = -aa[2, 0]
        aa[3, 2] = -aa[1, 0]
        aa[3, 3] = aa[0, 0]

    return numpy.real(aa)


@jitted
def dnka(omega, wvno, b, rho, cosp, rsinp, sinpr, cossv, rsinsv, sinsvr, ex, exa, iwat):
    """Dunkin's matrix for Rayleigh-wave."""
    ca = numpy.zeros((5, 5), dtype=numpy.complex_)
    wvno2 = wvno * wvno
    om2 = omega * omega

    if iwat == 1:
        # Fluid layer
        dfac = numpy.exp(-ex) if ex < 35.0 else 0.0
        ca[2, 2] = dfac
        ca[0, 0] = cosp
        ca[4, 4] = cosp
        ca[0, 1] = -rsinp / rho / om2
        ca[1, 0] = -rho * sinpr * om2
        ca[1, 1] = cosp
        ca[3, 3] = cosp
        ca[3, 4] = ca[0, 1]
        ca[4, 3] = ca[1, 0]
    else:
        a0 = numpy.exp(-exa) if exa < 60.0 else 0.0
        cpcq = cosp * cossv
        cpy = cosp * sinsvr
        cpz = cosp * rsinsv
        cqw = cossv * sinpr
        cqx = cossv * rsinp
        xy = rsinp * sinsvr
        xz = rsinp * rsinsv
        wy = sinpr * sinsvr
        wz = sinpr * rsinsv

        # Elastic layer
        rho2 = rho * rho
        gam = 2.0 * b * b * wvno2 / om2
        gam2 = gam * gam
        gamm1 = gam - 1.0
        gamm2 = gamm1 * gamm1
        cqww2 = cqw * wvno2
        cqxw2 = cqx / wvno2
        gg1 = gam * gamm1
        a0c = (2.0 + 0.0 * 1j) * (a0 + 0.0 * 1j - cpcq)
        xz2 = xz / wvno2
        gxz2 = gam * xz2
        g2xz2 = gam2 * xz2
        a0cgg1 = a0c * (gam + gamm1)
        wy2 = wy * wvno2
        g2wy2 = gamm2 * wy2
        g1wy2 = gamm1 * wy2

        # OK by symmetry
        temp = a0c * gg1 + g2xz2 + g2wy2
        ca[2, 2] = a0 + temp + temp
        ca[0, 0] = cpcq - temp
        ca[0, 1] = (-cqx + wvno2 * cpy) / rho / om2

        temp = (0.5 + 0.0 * 1j) * a0cgg1 + gxz2 + g1wy2
        ca[0, 2] = wvno * temp / rho / om2
        ca[0, 3] = (-cqww2 + cpz) / rho / om2

        temp = wvno2 * (a0c + wy2) + xz
        ca[0, 4] = -temp / rho2 / om2 / om2

        ca[1, 0] = (-gamm2 * cqw + gam2 * cpz / wvno2) * rho * om2
        ca[1, 1] = cpcq
        ca[1, 2] = (gamm1 * cqww2 - gam * cpz) / wvno
        ca[1, 3] = -wz
        ca[1, 4] = ca[0, 3]

        temp = (0.5 + 0.0 * 1j) * a0cgg1 * gg1 + gam2 * gxz2 + gamm2 * g1wy2
        ca[2, 0] = -(2.0 + 0.0 * 1j) * temp * rho * om2 / wvno
        ca[2, 1] = -wvno * (gam * cqxw2 - gamm1 * cpy) * (2.0 + 0.0 * 1j)
        ca[2, 3] = -2.0 * ca[1, 2]
        ca[2, 4] = -2.0 * ca[0, 2]

        ca[3, 0] = (-gam2 * cqxw2 + gamm2 * cpy) * rho * om2
        ca[3, 1] = -xy
        ca[3, 2] = -0.5 * ca[2, 1]
        ca[3, 3] = ca[1, 1]
        ca[3, 4] = ca[0, 1]

        temp = gamm2 * (a0c * gam2 + g2wy2) + gam2 * g2xz2
        ca[4, 0] = -rho2 * om2 * om2 * temp / wvno2
        ca[4, 1] = ca[3, 0]
        ca[4, 2] = -0.5 * ca[2, 0]
        ca[4, 3] = ca[1, 0]
        ca[4, 4] = ca[0, 0]

    return numpy.real(ca)


@jitted
def shup(omega, wvno, d, a, b, rho):
    """Find the elements of the Haskell matrix for Love-wave."""
    mmax = len(d)

    uu = numpy.zeros(mmax, dtype=numpy.float64)
    tt = numpy.zeros(mmax, dtype=numpy.float64)
    exl = numpy.zeros(mmax, dtype=numpy.float64)

    # Kludge for fluid core
    if b[-1] > 0.01:
        dpth = 0.0
        rb, mu, cosq, _, _, _, eexl = varl(mmax - 1, omega, wvno, dpth, b, rho)
        uu[-1] = 0.0
        tt[-1] = -mu * rb
    else:
        uu[-1] = 0.0
        tt[-1] = 1.0

    for i in range(mmax - 2, -1, -1):
        if b[i] > 0.01:
            dpth = d[i]
            rb, mu, cosq, _, y, z, eexl = varl(i, omega, wvno, dpth, b, rho)
            hl = hskl(i, b, mu, cosq, y, z)

            # We actually use A^-1 since we go from the bottom to top
            i1 = i + 1
            amp0 = hl[0, 0] * uu[i1] - hl[0, 1] * tt[i1]
            str0 = -hl[1, 0] * uu[i1] + hl[1, 1] * tt[i1]

            # Normalize
            rr = numpy.abs(amp0)
            ss = numpy.abs(str0)
            rr = max(rr, ss)
            if rr < 1.0e-30:
                rr = 1.0
            exl[i] = numpy.log(rr) + eexl
            uu[i] = amp0 / rr
            tt[i] = str0 / rr

    return uu, tt, exl


@jitted
def svup(omega, wvno, d, a, b, rho):
    """
    Find the values of the Dunkin vectors for Rayleigh-wave.

    Values are calculated at each layer boundaries from bottom layer upward.

    """
    mmax = len(d)

    ee = numpy.zeros(5, dtype=numpy.float64)
    cd = numpy.zeros((mmax, 5), dtype=numpy.float64)
    exe = numpy.zeros(mmax, dtype=numpy.float64)

    # Set up starting values for bottom halfspace
    gbr = evalg(mmax - 1, d, a, b, rho, wvno, omega)
    for i in range(5):
        cd[-1, i] = gbr[i]

    # Matrix multiplication from bottom layer upward
    wvno2 = wvno * wvno

    exsum = 0.0
    for m in range(mmax - 2, -1, -1):
        xka = omega / a[m]
        xkb = omega / b[m] if b[m] > 0.01 else 0.0
        rp = numpy.sqrt(wvno2 - xka * xka + 0.0 * 1j)
        rsv = numpy.sqrt(wvno2 - xkb * xkb + 0.0 * 1j)
        p = rp * d[m]
        q = rsv * d[m]

        iwat = int(b[m] < 0.01)
        cosp, cossv, rsinp, rsinsv, sinpr, sinsvr, pex, svex = varsv(
            p, q, rp, rsv, d[m], iwat
        )
        ca = dnka(
            omega,
            wvno,
            b[m],
            rho[m],
            cosp,
            rsinp,
            sinpr,
            cossv,
            rsinsv,
            sinsvr,
            pex,
            pex + svex,
            iwat,
        )

        for i in range(5):
            cr = 0.0
            for j in range(5):
                cr += cd[m + 1, j] * ca[j, i]
            ee[i] = cr

        ee, exn = normc(ee, 5)
        exsum += pex + svex + exn
        exe[m] = exsum
        for i in range(5):
            cd[m, i] = ee[i]

    return cd, exe


@jitted
def svdown(omega, wvno, d, a, b, rho):
    """
    Find the values of the Haskell vectors for Rayleigh-wave.

    Values are calculated at each layer boundaries from top layer downward.

    """
    mmax = len(d)

    aa0 = numpy.zeros(4, dtype=numpy.float64)
    vv = numpy.zeros((mmax, 4), dtype=numpy.float64)
    exa = numpy.zeros(mmax, dtype=numpy.float64)
    wvno2 = wvno * wvno

    # Initialize the top surface for the first column of the Haskell propagator
    vv[0, 0] = 1.0
    exa[0] = 0.0

    # Multiplication from top layer downward
    exsum = 0.0
    for m in range(mmax - 1):
        xka = omega / a[m]
        xkb = omega / b[m] if b[m] > 0.01 else 0.0
        rp = numpy.sqrt(wvno2 - xka * xka + 0.0 * 1j)
        rsv = numpy.sqrt(wvno2 - xkb * xkb + 0.0 * 1j)
        p = rp * d[m]
        q = rsv * d[m]

        iwat = int(b[m] < 0.01)
        cosp, cossv, rsinp, rsinsv, sinpr, sinsvr, pex, svex = varsv(
            p, q, rp, rsv, d[m], iwat
        )
        aa = hska(
            omega,
            wvno,
            b[m],
            rho[m],
            cosp,
            rsinp,
            sinpr,
            cossv,
            rsinsv,
            sinsvr,
            pex,
            svex,
            iwat,
        )

        for i in range(4):
            cc = 0.0
            for j in range(4):
                cc += aa[i, j] * vv[m, j]
            aa0[i] = cc

        aa0, ex2 = normc(aa0, 4)
        exsum += pex + ex2
        exa[m + 1] = exsum

        for i in range(4):
            vv[m + 1, i] = aa0[i]

    return vv, exa


@jitted
def shfunc(omega, wvno, d, a, b, rho):
    """
    Compute eigenfunctions for Love-wave.

    This evaluates the eigenfunctions by calling sub up.

    """
    mmax = len(d)

    uu, tt, exl = shup(omega, wvno, d, a, b, rho)

    ext = 0.0
    umax = uu[0]
    tt[0] = 0.0
    for i in range(1, mmax):
        if b[i] > 0.01:
            ext += exl[i - 1]
            fac = 1.0 / numpy.exp(ext) if ext < 80.0 else 0.0
            uu[i] *= fac
            tt[i] *= fac
        else:
            uu[i] = 0.0
            tt[i] = 0.0

        if numpy.abs(uu[i]) > numpy.abs(umax):
            umax = uu[i]

    if uu[0] != 0.0:
        umax = uu[0]

    if numpy.abs(umax) > 0.0:
        for i in range(mmax):
            if b[i] > 0.0:
                uu[i] /= umax
                tt[i] /= umax

    return uu, tt


@jitted
def svfunc(omega, wvno, d, a, b, rho):
    """
    Compute eigenfunctions for Rayleigh-wave.

    This combines the Haskell vector from sub down and Dunkin vector from sub up.

    """
    mmax = len(d)

    ur = numpy.zeros(mmax, dtype=numpy.float64)
    uz = numpy.zeros(mmax, dtype=numpy.float64)
    tz = numpy.zeros(mmax, dtype=numpy.float64)
    tr = numpy.zeros(mmax, dtype=numpy.float64)

    # Get compound matrix from bottom to top
    cd, exe = svup(omega, wvno, d, a, b, rho)
    vv, exa = svdown(omega, wvno, d, a, b, rho)

    # Get propagator from top to bottom
    f1213 = -cd[0, 1]
    ur[0] = cd[0, 2] / cd[0, 1]
    uz[0] = 1.0

    ext = 0.0
    for i in range(1, mmax):
        ext = exa[i] + exe[i] - exe[0]

        if -80.0 < ext < 80.0:
            cd1 = cd[i, 0]
            cd2 = cd[i, 1]
            cd3 = cd[i, 2]
            cd4 = -cd[i, 2]
            cd5 = cd[i, 3]
            cd6 = cd[i, 4]
            tz1 = -vv[i, 3]
            tz2 = -vv[i, 2]
            tz3 = vv[i, 1]
            tz4 = vv[i, 0]

            ur[i] = tz2 * cd6 - tz3 * cd5 + tz4 * cd4
            uz[i] = -tz1 * cd6 + tz3 * cd3 - tz4 * cd2
            tz[i] = tz1 * cd5 - tz2 * cd3 + tz4 * cd1
            tr[i] = -tz1 * cd4 + tz2 * cd2 - tz3 * cd1

            fac = numpy.exp(ext) / f1213
            ur[i] *= fac
            uz[i] *= fac
            tz[i] *= fac
            tr[i] *= fac
        else:
            ur[i] = 0.0
            uz[i] = 0.0
            tz[i] = 0.0
            tr[i] = 0.0

    # Correction for fluid layers on top if not all fluid
    if numpy.any(b > 0.01):
        jwat = 0
        for i in range(mmax):
            if b[i] < 0.01:
                jwat += 1
            else:
                break

        if jwat > 0.0:
            for i in range(jwat):
                ur[i] = 0.0
                tr[i] = 0.0

    return ur, uz, tz, tr


@jitted
def swegn96(t, d, a, b, rho, mode, ifunc=2, dc=0.005):
    """
    Get eigenfunctions for a given period and mode.

    Parameters
    ----------
    t : scalar
        Period (in s).
    d : array_like
        Layer thickness (in km).
    a : array_like
        Layer P-wave velocity (in km/s).
    b : array_like
        Layer S-wave velocity (in km/s).
    rho : array_like
        Layer density (in g/cm3).
    mode : int, optional, default 0
        Mode number (0 if fundamental).
    ifunc : int, optional, default 2
        Select wave type and algorithm for period equation:
         - 1: Love-wave (Thomson-Haskell method),
         - 2: Rayleigh-wave (Dunkin's matrix),
         - 3: Rayleigh-wave (fast delta matrix).
    dc : scalar, optional, default 0.005
        Phase velocity increment for root finding.

    Returns
    -------
    array_like
        Eigenfunctions.

    """
    mmax = len(d)

    # Compute eigenvalue (phase velocity)
    period = numpy.empty(1, dtype=numpy.float64)
    period[0] = t
    c = surf96(period, d, a, b, rho, mode, 0, ifunc, dc)
    omega = 2.0 * numpy.pi / t

    if c[0] > 0.0:
        wvno = omega / c[0]
    else:
        raise DispersionError("mode does not exist at given period")

    # Compute eigenfunctions
    if ifunc == 1:
        uu, tt = shfunc(omega, wvno, d, a, b, rho)

        egn = numpy.empty((mmax, 2), dtype=numpy.float64)
        for i in range(mmax):
            egn[i, 0] = uu[i]
            egn[i, 1] = tt[i]

    else:
        ur, uz, tz, tr = svfunc(omega, wvno, d, a, b, rho)

        egn = numpy.empty((mmax, 4), dtype=numpy.float64)
        for i in range(mmax):
            egn[i, 0] = ur[i]
            egn[i, 1] = uz[i]
            egn[i, 2] = tz[i]
            egn[i, 3] = tr[i]

    return egn
